"use strict";

// This wrapper for citeproc provides the high performance of incremental updates, without
// the added complexity.
//
// 1. It uses processCitationCluster() calls to insert and remove citation clusters
//    efficiently instead of re-rendering the entire document.
// 2. The API is conceptually simple, allowing clients to send the entire list of citation
//    clusters on each render, without worrying about maintaining state.
//
// Incremental rendering works by performing a diff between the new and old document states,
// and generating the minimal list of processCitationCluster() calls required to transform
// the previous list of clusters to the list.
//
// A full re-render can be triggered by either changing the citation style, or specifiying
// the {reset: true} option.
//
// If reference items are specified, but no citationClusters are specified, a bibliography
// only render mode is triggered which is slightly faster. (Note: in future it would be
// nice to add the ability to specify a mix of cited and uncited references, which is
// supported by citeproc-js.)
//
// Documentation for citeproc-js: http://gsl-nagoya-u.net/http/pub/citeproc-doc.html

var CiteprocReact = function (options) {
	var that = this;

	options = _.pick(options, [
			'citeprocSys' // object - configuration for citeproc
		]) || {};

	this.clusterList;         // ClusterList - for the current render
	this.previousClusterList; // ClusterList - for the previous render
	this.style;               // string      - the CSL style
	this.locale;              // string      - the locale ("en-US", "en-GB", etc.)
	this.citeproc;            // CSL.Engine  - citeproc-js instance

	this.formattedCitations;           // list   - inline citations from citeproc in order of
	                                   //          appearance in document
	this.formattedBibliography;        // list   - ordered bibliography entries from citeproc
	this.formattedBibliographyOptions; // object - options from citeproc
	this.renderInfo;                   // object - information about the previous render

	this.citeprocSys = options.citeprocSys;	// configuration for citeproc

	this.citeprocSys.retrieveItem = function (itemID) {
		var result = that.clusterList.getItem(itemID);
		if (!result && that.previousClusterList) {
			result = that.previousClusterList.getItem(itemID);
		}
		return result;
	};
};

CiteprocReact.prototype.render = function (renderOptions) {
	renderOptions = _.pick(renderOptions, [
			'style',             // string  - the CSL style (required)
			'locale',	     // string - e.g. "en-US", "en-GB", ...
			'citationClusters',  // list    - the list of citationClusters (see unitTests.js for examples)
			'items',             // object  - map of itemIDs to items (see unitTests.js for examples)
			                     //           to render a bibliography only, without in-line citations,
			                     //           use ordered IDs in the following format:
			                     //           "ITEM-1", "ITEM-2", ..., "ITEM-42", etc...
			'reset',             // boolean - if true, a full re-render is performed
			'enumerateCitations',// boolean - e.g. for IEEE: true (default) -> "[1]"
			                     //                          false          -> "citation-number"
			'bibliographyOutputFormat'// string - "rtf" or "html"
		]);

	var startTime = new Date().getTime();
	this._init(renderOptions);

	this.previousClusterList = this.clusterList;

	if (!renderOptions.citationClusters || renderOptions.citationClusters.length === 0) {
		this._bibliographyOnlyCiteprocRender(renderOptions);
	} else {
		this.clusterList = new ClusterList({
			citationClusters: renderOptions.citationClusters,
			items: renderOptions.items
		});
		if (renderOptions.reset || !this.previousClusterList) {
			this._fullCiteprocRender(renderOptions);
		} else {
			this._incrementalCiteprocRender(renderOptions);
		}
	}

	this.renderInfo.timeTaken = new Date().getTime() - startTime;

	var result = {
		formattedCitations: _.pluck(this.formattedCitations, 'formattedCitation'),
		formattedBibliography: this.formattedBibliography,
		formattedBibliographyOptions: this.formattedBibliographyOptions,
		info: this.renderInfo
	};

	return result;
};

CiteprocReact.prototype._init = function (renderOptions) {
	if (renderOptions.reset || !this.citeproc || this.style != renderOptions.style ||
			this.locale != renderOptions.locale || renderOptions.citationClusters.length === 0) {
		delete this.clusterList;
		delete this.previousClusterList;
		
		if (this.citeproc && this.style === renderOptions.style && this.locale === renderOptions.locale) {
			this.citeproc.restoreProcessorState([]);
		} else {
			this.style = renderOptions.style;
			this.locale = renderOptions.locale;
			this.citeproc = new CSL.Engine(this.citeprocSys,
					this.style,
					this.locale,
					true /* uses the specified locale even if the style has a default locale */
			);
			this.citeproc.setAbbreviations("default");
		}

		// Clear the abbreviation cache
		this.citeproc.transform.abbrevs = new CSL.AbbreviationSegments();
	}
};

// unlike the "full" and "incremental" renders, this expects the reference items
// to have ids in the form "ITEM-1", "ITEM-2", etc...
CiteprocReact.prototype._bibliographyOnlyCiteprocRender = function (renderOptions) {
	this.clusterList = {
		getItem: function (itemID) {
			return renderOptions.items[itemID];
		}
	};

	var orderedItemIDs = [];

	var itemRegExp = new RegExp("^ITEM-([0-9]+)$");

	// order items
	_.each(renderOptions.items, function (item) {
		var match = itemRegExp.exec(item.id);
		if (!match) {
			throw new Error('Items must have ID in the form "ITEM-XX" for bibliography only render');
		}
		orderedItemIDs[parseInt(match[1]) - 1] = item.id;
	});

	this.citeproc.updateItems([]);
	this.citeproc.updateUncitedItems(orderedItemIDs);

	this._renderBibliography(renderOptions);
	this.formattedCitations = [];

	// cleanup
	this.citeproc.updateUncitedItems([]);
	delete this.clusterList;
	delete this.previousClusterList;

	this.renderInfo = {
		type: 'bibliographyOnly'
	};
};

CiteprocReact.prototype._fullCiteprocRender = function (renderOptions) {
	var that = this;
	
	this.citeproc.setOutputFormat("html");

	this.citeproc.updateItems(this.clusterList.getOrderedItemIDs());
	var citeprocFormattedCitations = this.citeproc.rebuildProcessorState(this.clusterList.getCiteprocClusters());

	if (this.clusterList.clusters.length !== citeprocFormattedCitations.length) {
		throw new Error("formatted citations don't match input citations");
	}
	this.formattedCitations = _.map(citeprocFormattedCitations, function (citeprocResultCluster, index) {
		return {
			formattedCitation: citeprocResultCluster[2],
			citationID: that.clusterList.clusters[index].cluster.citationID
		};
	});
	this._renderBibliography(renderOptions);
	this.renderInfo = {
		type: 'full'
	};
};

CiteprocReact.prototype._incrementalCiteprocRender = function (renderOptions) {
	var that = this;
	
	this.citeproc.setOutputFormat("html");

	var minimalChanges = this.clusterList.minimalCiteprocChanges(this.previousClusterList);
	_.each(minimalChanges, function (change) {
		var formattedCitationChanges =
			that.citeproc.processCitationCluster(change.citation, change.citationsPre, change.citationsPost)[1];

		var preCitationIDs = _.pluck(change.citationsPre, 0);
		var postCitationIDs = _.pluck(change.citationsPost, 0);

		var newFormattedCitations = _.filter(that.formattedCitations, function (formattedCitation) {
			return _.contains(preCitationIDs, formattedCitation.citationID);
		});
		newFormattedCitations.push({citationID: change.citation.citationID});
		newFormattedCitations = newFormattedCitations.concat(_.filter(that.formattedCitations, function (formattedCitation) {
			return _.contains(postCitationIDs, formattedCitation.citationID);
		}));
		that.formattedCitations = newFormattedCitations;

		_.each(formattedCitationChanges, function (formattedCitationChange) {
			that.formattedCitations[formattedCitationChange[0]].formattedCitation = formattedCitationChange[1];
		});
	});

	this._renderBibliography(renderOptions);

	this.renderInfo = {
		type: 'incremental',
		numChanges: minimalChanges.length
	};
};

CiteprocReact.prototype._renderBibliography = function (renderOptions) {
	var citationNumberString;
	
	this.citeproc.setOutputFormat(renderOptions.bibliographyOutputFormat);

	if (!renderOptions.enumerateCitations) {
		citationNumberString = "citation-number";
	}
	
	var citeprocFormattedBibliography = this.citeproc.makeBibliography(citationNumberString);
	this.formattedBibliographyOptions = citeprocFormattedBibliography[0];
	this.formattedBibliography = citeprocFormattedBibliography[1];
};

