"use strict";

// This represents a list of citation clusters and can generate a list of citeproc-js
// commands to transform one list into another.
//
// In order to compare two ClusterLists, we perfom a line-based diff on them, where each
// line is a citation cluster JSON string.
//
// Citation clusters contain a list of reference item ID. For the diff to be useful, equivalent
// reference item IDs must represent the exact same reference metadata.
// This means that if the same reference is edited between renders, its ID must change, and
// it's treated as a completely new reference.
//
// To be sure of this, we don't trust any IDs provided by the client and generate each item
// ID by taking an MD5 hash of its metadata.
//
// This has the side effect of treating two references which may be distinct in a user's
// reference manager as the same reference item if they have the exact same metadata.

var ClusterList = function (options) {
	this.options = _.pick(options, [
			"citationClusters", // list   - citation clusters
			"items"	            // object - map of given itemIDs to items
		]);

	this.itemsByID = {};             // map of all new itemIDs (based on MD5 hashes) to items 
	this.nextCiteprocCitationID = 0; // counter which increments as each new citeproc-ready
	                                 // citation cluster is generated

	this.clusters = [];  // list of normalised citation clusters
	this.hashCache = {}; // cache of MD5 hash results

	this._prepareCitationClusters();
};

// Normalises all citation clusters, ready both for diffing and sending to citeproc
ClusterList.prototype._prepareCitationClusters = function () {
	var that = this;

	_.each(deepCopy(this.options.citationClusters), function (cluster) {
		_.each(cluster.citationItems, function (item) {
			if (!item.item) {
				item.item = that.options.items[item.id];
			}
			var preparedItem = that._prepareItem(item.item);
			item.id = preparedItem.id;
			delete item.item;
		});
		delete cluster.citationID;
		that.clusters.push({
			clusterString: JSON.stringify(_.pick(cluster, ['citationItems', 'properties'])),
			cluster: cluster
		});
	});
};

// Caching MD5 hashes offers speedup for documents with repeated reference items
ClusterList.prototype.MD5 = function (input) {
	var val = this.hashCache[input];
	if (val) {
		return val;
	}
	val = CryptoJS.MD5(input).toString();
	this.hashCache[input] = val;
	return val;
};

// Prepares an item. Gives it an ID which is a hash (MD5) of it's own contents
// and adds it to itemsByID
ClusterList.prototype._prepareItem = function (item) {
	var originalID = item.id;
	item.id = this.MD5(JSON.stringify(_.omit(item, 'id')));
	this.itemsByID[item.id] = item;
	return item;
};

// Gets an item based on the new hash-based ID, not IDs given by the client
ClusterList.prototype.getItem = function (itemID) {
	return this.itemsByID[itemID];
};

ClusterList.prototype._diffClusterLists = function (clusters1, clusters2) {
	return diffStringLists(
		_.pluck(clusters1, 'clusterString'),
		_.pluck(clusters2, 'clusterString'));
};

ClusterList.prototype.getCiteprocClusters = function () {
	var that = this;
	var result = _.pluck(this.clusters, 'cluster');
	
	// populate with citationIDs if they don't already exist
	_.each(result, function (cluster) {
		if (!cluster.citationID) {
			cluster.citationID = that._newCiteprocID();
		}
		cluster.properties = cluster.properties || {};
	});

	return deepCopy(result);
};

// requires citeprocClusters() to have been called already
ClusterList.prototype.getOrderedItemIDs = function () {
	var orderedItemIDs = [],
		insertedItems = {};

	_.each(this.clusters, function (cluster) {
		_.each(cluster.cluster.citationItems, function (item) {
			if (!insertedItems[item.id]) {
				orderedItemIDs.push(item.id);
				insertedItems[item.id] = true;
			}
		});
	});
	
	return orderedItemIDs;
};

ClusterList.prototype._newCiteprocID = function () {
	var citationID = "CITATION-" + this.nextCiteprocCitationID;
	this.nextCiteprocCitationID++;
	return citationID;
};

ClusterList.prototype.getCluster = function (citationID) {
	return _.find(this.clusters, function (cluster) {
		return cluster.cluster.citationID === citationID;
	});
};

ClusterList.prototype.minimalCiteprocChanges = function (previousClusterList) {
	var that = this;

	// start indexing where the previous list left off
	this.nextCiteprocCitationID = previousClusterList.nextCiteprocCitationID;

	var diffs = this._diffClusterLists(previousClusterList.clusters, this.clusters);

	var citeprocChanges = [],
		originalIndex = -1,
		currentIndex = -1;

	var citationsPre = [],
		itemsDeleted = false;

	var addCiteprocChange = function (newCluster) {
		// next items
		var citationsPost = [];

		if (originalIndex + 1 < previousClusterList.clusters.length) {
			citationsPost = _.map(_.range(originalIndex + 1, previousClusterList.clusters.length), function (index) {
				return [previousClusterList.clusters[index].cluster.citationID, 0 /* footnote index */];
			});
		}

		if (!newCluster) {
			// citeproc NEEDS us to specify a new cluster, so grab an existing one from either pre or post
			if (citationsPre.length > 0) {
				newCluster = that.getCluster(citationsPre.pop()[0]).cluster;
			} else if (citationsPost.length > 0) {
				newCluster = previousClusterList.getCluster(citationsPost.splice(0, 1)).cluster;
			} else {
				console.error('Trying to delete all citations - not supported!');
			}
		}

		newCluster.properties = newCluster.properties || {};

		citeprocChanges.push({
			citation: newCluster,
			citationsPre: _.clone(citationsPre),
			citationsPost: citationsPost
		});
		citationsPre.push([newCluster.citationID, 0 /* footnote index */]);

		itemsDeleted = false;
	};

	_.each(diffs, function (diff) {
		var citeprocChange;

		switch (diff.action) {
			case "equal":
				_.each(diff.items, function (clusterString) {
					currentIndex++;
					originalIndex++;
					var citationID = previousClusterList.clusters[originalIndex].cluster.citationID;
					that.clusters[currentIndex].cluster.citationID = citationID;
					citationsPre.push([citationID, 0 /* footnote index */]);
				});
				break;
			case "remove":
				// don't add to new list
				originalIndex += diff.items.length;
				itemsDeleted = true;
				break;
			case "insert":
				_.each(diff.items, function (clusterString) {
					currentIndex++;

					var newCluster = that.clusters[currentIndex].cluster;
					newCluster.citationID = that._newCiteprocID();
					addCiteprocChange(newCluster);

					itemsDeleted = false;
				});
				break;
			default:
				throw new Error("Unexpected diff action: " + diff.action);
		}
	});

	if (itemsDeleted) {
		addCiteprocChange();
	}
	return citeprocChanges;
};
