"use strict";

// Test utility functions
var fetchTextSync = function (url) {
	var xhr = new XMLHttpRequest();
	xhr.open('GET', url, false);
	xhr.send(null);
	return xhr.responseText;
};

var newCiteprocReact = function() {
	var citeprocReact = new CiteprocReact({
		citeprocSys: {
			retrieveLocale: function(lang) {
			    return fetchTextSync('locales-en-US.xml');
			}
		}
	});
	return citeprocReact;
}

var randomSeed = 0;
var psuedoRandomNumber = function (x) {
	if (!x) {
		x = randomSeed++;
	}
	return Math.abs(Math.sin(1000 * x) * 1000 % 1);
};

var psuedoRandomInteger = function (max, x) {
	if (!x) {
		x = randomSeed++;
	}
	return Math.round(max * psuedoRandomNumber(x));
};

var testCitationItem = function (number, extraProperties) {
	return _.extend({
		id: 'item-' + number,
		item: {
			id: 'item-' + number,
			title: 'Article ' + number,
			author: ['Steve', 'Rob', 'Carles', 'George'][number % 4] + ' ' +
				['Ridout', 'Knight', 'Pina', 'Vafiadis', 'Kartvelishvili'][number % 5],
			issued: {
			   "date-parts": [[(1980 + Math.round(35 * psuedoRandomNumber(number)))]]
			}
		}
	}, extraProperties);
};

// Simplify citeproc bibliography entry HTML output
CSL.Output.Formats["html"]["@bibliography/entry"] = function (state, str) {
	return str;
};

QUnit.module("ClusterList");
QUnit.test("Normalise items", function (assert) {
	var testItems = [
		{
			title: 'hello',
			id: 'irrelevant'
		},
		{
			title: 'hello',
			id: 'this will be replaced'
		},
		{
			title: 'hello',
			author: 'Steve',
			id: 'irrelevant'
		}
	];

	var clusterList = new ClusterList({citationClusters: testItems});

	var normalisedTestItems = _.map(testItems, function (item) {
		return clusterList._prepareItem(item);
	});

	assert.ok(normalisedTestItems[0].id.length > 0);
	assert.deepEqual(normalisedTestItems[0], normalisedTestItems[1]);
	assert.notDeepEqual(normalisedTestItems[0], normalisedTestItems[2]);
});

QUnit.test("Prepare citation clusters", function (assert) {
	var testClusters = [
		{
			citationID: 'irrelevant 1',
			citationItems: [testCitationItem(1, {'author-only': true})]
		},
		{
			citationID: 'irrelevant 2',
			citationItems: [testCitationItem(1), testCitationItem(2)]
		},
		{
			citationID: 'irrelevant 3',
			citationItems: [testCitationItem(1, {'author-only': true})]
		}
	];

	var clusterList = new ClusterList({citationClusters: testClusters});

	assert.equal(clusterList.clusters[0].cluster.citationItems[0].id, "7e145d097ff3a93d2394ad683cbf5739");
	assert.equal(clusterList.clusters[1].cluster.citationItems[0].id, clusterList.clusters[0].cluster.citationItems[0].id);
	assert.notEqual(clusterList.clusters[1].cluster.citationItems[1].id, clusterList.clusters[0].cluster.citationItems[0].id);
	
	assert.equal(
		clusterList.getItem(clusterList.clusters[0].cluster.citationItems[0].id).title,
		testCitationItem(1).item.title, 'getItem()');
	assert.equal(clusterList.clusters[0].cluster.citationItems[0]['author-only'], true);

	assert.notEqual(clusterList.clusters[0].clusterString, clusterList.clusters[1].clusterString, "non-equivalent clusters");
	assert.equal(clusterList.clusters[0].clusterString, clusterList.clusters[2].clusterString, "equivalent clusters");

	var citeprocClusters = clusterList.getCiteprocClusters();

	assert.equal(citeprocClusters[0].citationID, "CITATION-0", "citeproc citationID");
	assert.equal(citeprocClusters[1].citationID, "CITATION-1", "citeproc citationID");
	assert.equal(citeprocClusters[2].citationID, "CITATION-2", "citeproc citationID");

	assert.equal(citeprocClusters[0].citationItems[0].id, clusterList.clusters[0].cluster.citationItems[0].id,
		"citeproc itemID");
});

QUnit.test("Diff string lists", function (assert) {
	var list1 = [
			"one",
			"two",
			"three",
			"four",
			"five"
		],
		list2 = [
			"one",
			"two",
			"new!",
			"four"
		];

	var diffs = diffStringLists(list1, list2);

	assert.deepEqual(diffs, [
		{action: "equal",  items: ["one", "two"]},
		{action: "remove", items: ["three"]},
		{action: "insert", items: ["new!"]},
		{action: "equal",  items: ["four"]},
		{action: "remove", items: ["five"]}
	]);
});

QUnit.test("Diff cluster change", function (assert) {
	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(4)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			}
		],
		testClusters2 = deepCopy(testClusters);
	
	var clusterList1 = new ClusterList({citationClusters: testClusters});

	// change 2nd cluster
	testClusters2[1] = {
		citationItems: [
			testCitationItem(3)
		]
	};

	// remove last cluster
	testClusters2.pop();

	var clusterList2 = new ClusterList({citationClusters: testClusters2});
	var diffs = ClusterList.prototype._diffClusterLists(clusterList1.clusters, clusterList2.clusters);

	assert.deepEqual(_.pluck(diffs, 'action'), ["equal", "remove", "insert", "equal", "remove"]);

	// render clusterList 1 (creates citeproc citationIDs)
	var citeprocClusters1 = clusterList1.getCiteprocClusters();

	assert.equal(citeprocClusters1[0].citationID, "CITATION-0", "citationID");
	assert.equal(citeprocClusters1[1].citationID, "CITATION-1", "citationID");
});

QUnit.test("Diff cluster change: removal of penultimate cluster", function (assert) {
	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(4)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			}
		],
		testClusters2 = deepCopy(testClusters);

	testClusters2.splice(2, 1);

	var clusterList1 = new ClusterList({citationClusters: testClusters});
	clusterList1.getCiteprocClusters();

	assert.equal(clusterList1.clusters.length, 4);
	assert.equal(clusterList1.clusters[3].cluster.citationID, "CITATION-3");

	var clusterList2 = new ClusterList({citationClusters: testClusters2});

	var changes = clusterList2.minimalCiteprocChanges(clusterList1);

	assert.equal(changes.length, 1, "one change");
	var change = changes.pop();
	assert.equal(change.citation.citationID, "CITATION-3");
	assert.deepEqual(_.pluck(change.citationsPre, 0), ["CITATION-0", "CITATION-1"]);
	assert.deepEqual(_.pluck(change.citationsPost, 0), []);
});

QUnit.test("Diff cluster change: removal of last cluster", function (assert) {
	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(4)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			}
		],
		testClusters2 = deepCopy(testClusters);

	testClusters2.pop();

	var clusterList1 = new ClusterList({citationClusters: testClusters});
	clusterList1.getCiteprocClusters();

	assert.equal(clusterList1.clusters.length, 4);
	assert.equal(clusterList1.clusters[3].cluster.citationID, "CITATION-3");

	var clusterList2 = new ClusterList({citationClusters: testClusters2});
	var changes = clusterList2.minimalCiteprocChanges(clusterList1);

	assert.equal(changes.length, 1, "one change");
	var change = changes.pop();
	assert.equal(change.citation.citationID, "CITATION-2");
	assert.deepEqual(_.pluck(change.citationsPre, 0), ["CITATION-0", "CITATION-1"]);
	assert.deepEqual(_.pluck(change.citationsPost, 0), []);
});

QUnit.test("Diff cluster change: append cluster", function (assert) {
	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)]
			}
		],
		testClusters2 = deepCopy(testClusters);

	testClusters2.push({citationItems: [testCitationItem(3)]});

	var clusterList1 = new ClusterList({citationClusters: testClusters});
	clusterList1.getCiteprocClusters();

	var clusterList2 = new ClusterList({citationClusters: testClusters2});
	var changes = clusterList2.minimalCiteprocChanges(clusterList1);

	assert.equal(changes.length, 1, "one change");
	var change = changes.pop();
	assert.equal(change.citation.citationID, "CITATION-2");
	assert.deepEqual(_.pluck(change.citationsPre, 0), ["CITATION-0", "CITATION-1"]);
	assert.deepEqual(_.pluck(change.citationsPost, 0), []);
});

QUnit.test("Diff cluster change: prepend cluster", function (assert) {
	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)]
			}
		],
		testClusters2 = deepCopy(testClusters);

	testClusters2.splice(0, 0, {citationItems: [testCitationItem(3)]});

	var clusterList1 = new ClusterList({citationClusters: testClusters});
	clusterList1.getCiteprocClusters();

	var clusterList2 = new ClusterList({citationClusters: testClusters2});
	var changes = clusterList2.minimalCiteprocChanges(clusterList1);

	assert.equal(changes.length, 1, "one change");
	var change = changes.pop();
	assert.equal(change.citation.citationID, "CITATION-2");
	assert.deepEqual(_.pluck(change.citationsPre, 0), []);
	assert.deepEqual(_.pluck(change.citationsPost, 0), ["CITATION-0", "CITATION-1"]);
});

QUnit.module("CiteprocReact");
var doRenderingTest = function (assert, alwaysReset) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = [
			{
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationItems: [testCitationItem(1), testCitationItem(2)]
			},
			{
				citationItems: [testCitationItem(3, {'author-only': true})]
			}
		];

	var result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		reset: alwaysReset,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 3, 'formatted citations length');
	assert.equal(result.formattedCitations[0], "Rob Knight");
	assert.equal(result.formattedCitations[1], "(Carles Pina, 1981; Rob Knight, 2011)");
	assert.equal(result.formattedCitations[2], "George Vafiadis");

	assert.equal(result.formattedBibliography.length, 3, 'formatted bibliography length');
	assert.equal(result.formattedBibliography[0], "Carles Pina. (1981). Article 2.");
	assert.equal(result.formattedBibliography[1], "George Vafiadis. (1987). Article 3.");
	assert.equal(result.formattedBibliography[2], "Rob Knight. (2011). Article 1.");

	// Alter by appending a citation cluster
	testClusters.push({citationItems: [testCitationItem(4), testCitationItem(2)]});

	// Re-render
	result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		reset: alwaysReset,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 4, 'formatted citations length');
	assert.equal(result.formattedCitations[0], "Rob Knight");
	assert.equal(result.formattedCitations[1], "(Carles Pina, 1981; Rob Knight, 2011)");
	assert.equal(result.formattedCitations[2], "George Vafiadis");
	assert.equal(result.formattedCitations[3], "(Carles Pina, 1981; Steve Kartvelishvili, 1998)");

	assert.equal(result.formattedBibliography.length, 4, 'formatted bibliography length');
	assert.equal(result.formattedBibliography[0], "Carles Pina. (1981). Article 2.");
	assert.equal(result.formattedBibliography[1], "George Vafiadis. (1987). Article 3.");
	assert.equal(result.formattedBibliography[2], "Rob Knight. (2011). Article 1.");
	assert.equal(result.formattedBibliography[3], "Steve Kartvelishvili. (1998). Article 4.");

	// Alter by editing a citation cluster
	testClusters[1] = {citationItems: [testCitationItem(5), testCitationItem(2)]};

	// Re-render
	result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		reset: alwaysReset,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 4, 'edit cluster: formatted citations length');
	assert.equal(result.formattedCitations[0], "Rob Knight");
	assert.equal(result.formattedCitations[1], "(Carles Pina, 1981; Rob Ridout, 2014)");
	assert.equal(result.formattedCitations[2], "George Vafiadis");
	assert.equal(result.formattedCitations[3], "(Carles Pina, 1981; Steve Kartvelishvili, 1998)");

	assert.equal(result.formattedBibliography.length, 5, 'edit cluster: formatted bibliography length');
	assert.equal(result.formattedBibliography[0], "Carles Pina. (1981). Article 2.");
	assert.equal(result.formattedBibliography[1], "George Vafiadis. (1987). Article 3.");
	assert.equal(result.formattedBibliography[2], "Rob Knight. (2011). Article 1.");
	assert.equal(result.formattedBibliography[3], "Rob Ridout. (2014). Article 5.");
	assert.equal(result.formattedBibliography[4], "Steve Kartvelishvili. (1998). Article 4.");

	// Completely new list
	result = citeprocReact.render({
		style: style,
		citationClusters: [
			{citationItems: [testCitationItem(6)]},
			{citationItems: [testCitationItem(7)]},
			{citationItems: [testCitationItem(8), testCitationItem(9)]},
		],
		reset: alwaysReset,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 3, 'change all: formatted citations length');
	assert.equal(result.formattedCitations[0], "(Carles Knight, 2005)");
	assert.equal(result.formattedCitations[1], "(George Pina, 2011)");
	assert.equal(result.formattedCitations[2], "(Rob Kartvelishvili, 1996; Steve Vafiadis, 2010)");

	assert.equal(result.formattedBibliography.length, 4, 'change all: formatted bibliography length');
	assert.equal(result.formattedBibliography[0], "Carles Knight. (2005). Article 6.");
	assert.equal(result.formattedBibliography[1], "George Pina. (2011). Article 7.");
	assert.equal(result.formattedBibliography[2], "Rob Kartvelishvili. (1996). Article 9.");
	assert.equal(result.formattedBibliography[3], "Steve Vafiadis. (2010). Article 8.");

	// Remove all clusters
	result = citeprocReact.render({
		style: style,
		citationClusters: [],
		reset: alwaysReset,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 0, 'remove all citation clusters');
	assert.equal(result.formattedBibliography.length, 0, 'remove all citation clusters');
};
QUnit.test("Render small test documents: always reset", function (assert) {
	doRenderingTest(assert, true);
});

QUnit.test("Render small test documents: incremental updates", function (assert) {
	doRenderingTest(assert, false);
});

QUnit.test("incremental render - add 2 identical new items", function (assert) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = [
			{
				citationItems: [testCitationItem(1, {'author-only': true})]
			}
		];
	var testItems = {};
	testItems[testCitationItem(1).id] = testCitationItem(1).item;

	var result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 1);
	assert.equal(result.formattedCitations[0], "Rob Knight");

	assert.equal(result.formattedBibliography.length, 1);
	assert.equal(result.formattedBibliography[0], "Rob Knight. (2011). Article 1.");

	testClusters.push({citationItems: [{ id: 'new-item-id' }]});
	testClusters.push({citationItems: [{ id: 'new-item-id' }]});

	testItems['new-item-id'] = {title: 'new title', author: 'new author'};

	result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 3);
	assert.equal(result.formattedCitations[0], "Rob Knight");
	assert.equal(result.formattedCitations[1], "(new author, n.d.)");
	assert.equal(result.formattedCitations[2], "(new author, n.d.)");

	assert.equal(result.formattedBibliography.length, 2);
	assert.equal(result.formattedBibliography[0], "new author. (n.d.). new title.");
	assert.equal(result.formattedBibliography[1], "Rob Knight. (2011). Article 1.");
});

QUnit.test("incremental render - different citation IDs", function (assert) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = [
			{
				citationID: 'test-id-1',
				citationItems: [testCitationItem(1, {'author-only': true})]
			}
		];
	var testItems = {};
	testItems[testCitationItem(1).id] = testCitationItem(1).item;

	var result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		bibliographyOutputFormat: "html"
	});

	// same as first but with different id
	testClusters[0].citationID = "changed-id";

	var result2 = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		bibliographyOutputFormat: "html"
	});
	
	assert.deepEqual(_.omit(result, 'info'), _.omit(result2, 'info'));
});

QUnit.test("render empty list", function (assert) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	var result = citeprocReact.render({
		style: style,
		citationClusters: [],
		bibliographyOutputFormat: "html"
	});

	assert.equal(result.formattedCitations.length, 0);
	assert.equal(result.formattedBibliography.length, 0);

	var result2 = citeprocReact.render({
		style: style,
		citationClusters: [],
		bibliographyOutputFormat: "html"
	});

	assert.equal(result2.formattedCitations.length, 0);
	assert.equal(result2.formattedBibliography.length, 0);
});

QUnit.test("render unicode separator", function (assert) {
	var style = fetchTextSync('ama-separator-test.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = [
		{
			citationItems: [
				{
					item: {
						title: "A Title",
						author: [
							{
								family: "Gediminas",
								given: "Adomavicius"
							},
							{
								family: "Ramesh",
								given: "Sandaranarayanan"
							}
						]
					}
				}
			]
		}
	];

	var result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		bibliographyOutputFormat: "html"
	});

	var char1 = String.fromCharCode(237),
		char2 = String.fromCharCode(224);

	assert.ok(result.formattedBibliography.join("").indexOf(char1) !== -1, "contains " + char1);
	assert.ok(result.formattedBibliography.join("").indexOf(char2) !== -1, "contains " + char2);
});

QUnit.test("change styles", function (assert) {
	var citeprocReact = newCiteprocReact();

	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(4)]
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})]
			}
		];
	
	_.each(['apa', 'ieee', 'nature'], function (styleName) {
		var startTime = new Date().getTime();
		var result = citeprocReact.render({
			style: fetchTextSync('external/' + styleName + '.csl'),
			citationClusters: testClusters,
			bibliographyOutputFormat: "html"
		});
		var fullAPARenderTime = new Date().getTime() - startTime;
		assert.equal(result.info.type, 'full', 'full render');

		var startTime = new Date().getTime();
		var result = citeprocReact.render({
			style: fetchTextSync('external/' + styleName + '.csl'),
			citationClusters: testClusters,
			bibliographyOutputFormat: "html"
		});
		var incrementalAPARenderTime = new Date().getTime() - startTime;
		assert.equal(result.info.type, 'incremental', 'incremental render');
		assert.ok(incrementalAPARenderTime < fullAPARenderTime, 'incremental faster than full render');
	});
});

QUnit.test("extra (e.g. mendeley) citation cluster fields", function (assert) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = [
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1, {'author-only': true})],
				mendeley: 'look away'
			},
			{
				citationID: 'irrelevant',
				citationItems: [testCitationItem(1), testCitationItem(2)],
				mendeley: 'ignore this',
				arbiraryOtherField: "pay no heed"
			}
		];

	var result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		bibliographyOutputFormat: "html"
	});
	assert.equal(result.info.type, 'full');

	testClusters[0].mendeley = 'new';
	testClusters[1].mendeley = 'new';
	
	result = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		bibliographyOutputFormat: "html"
	});
	assert.equal(result.info.type, 'incremental');
	assert.equal(result.info.numChanges, 0); 
});

QUnit.test("Render bibliography only", function (assert) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	var originalTestItems = _.map(_.range(5), testCitationItem);
	var testItems = {};

	var index = 1;
	// set item keys correctly for bibliography test
	_.each(originalTestItems, function (item) {
		var id = "ITEM-" + index;
		testItems[id] = item.item;
		testItems[id].id = id;
		index++;
	});
	var testClusters = _.map(testItems, function (item) {
			return {
				citationItems: [item]
			}
		});

	var fullRenderResult = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: deepCopy(testItems),
		bibliographyOutputFormat: "html"
	});
	var bibliographyOnlyResult = citeprocReact.render({
		style: style,
		citationClusters: [],
		items: deepCopy(testItems),
		reset: true,
		bibliographyOutputFormat: "html"
	});
	assert.deepEqual(bibliographyOnlyResult.formattedBibliography, fullRenderResult.formattedBibliography);
});

var failedTests = [];

QUnit.testDone(function (data) {
	if (data.failed > 0) {
		failedTests.push(data.name);
	}
});

QUnit.done(function (data) {
	if (typeof(MendeleyDesktop) !== "undefined") {
		MendeleyDesktop.unitTestsFinished(data.passed, data.total, failedTests.join('\n'));
	}
});

// ---- slow tests start here ----

/*
var largeDocumentRenderingTest = function (styleName, assert, alwaysReset) {
	var style = fetchTextSync('external/' + styleName + '.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = JSON.parse(fetchTextSync('testCitations.json'));
	var testItems = JSON.parse(fetchTextSync('testItems.json'));

	// set item keys correctly for bibliography test
	_.each(testItems, function (item, key) {
		item.id = key;
	});

	var result1 = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		reset: alwaysReset
	});

	var result2 = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		reset: alwaysReset
	});

	assert.deepEqual(_.omit(result1, 'info'), _.omit(result2, 'info'));

	// Alter by appending and prepending a citation cluster
	testClusters.splice(0, 0, {citationItems: [testCitationItem(1), testCitationItem(2)]});

	console.time('large document prepend. alwaysReset=' + alwaysReset + ', style=' + styleName);
	var result3 = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		reset: alwaysReset
	});
	console.timeEnd('large document prepend. alwaysReset=' + alwaysReset + ', style=' + styleName);

	testClusters.push({citationItems: [testCitationItem(3), testCitationItem(4)]});

	console.time('large document append. alwaysReset=' + alwaysReset + ', style=' + styleName);
	var result4 = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		reset: alwaysReset
	});
	console.timeEnd('large document append. alwaysReset=' + alwaysReset + ', style=' + styleName);

	assert.equal(result4.formattedCitations.length, result2.formattedCitations.length + 2);
	assert.equal(result4.formattedBibliography.length, result2.formattedBibliography.length + 4);
};

QUnit.test("Render large document: always reset", function (assert) {
	console.log('---- render large - full render');
	largeDocumentRenderingTest('apa', assert, true);
});

QUnit.test("Render large document: incremental updates", function (assert) {
	console.log('---- render large - incremental render');
	largeDocumentRenderingTest('apa', assert, false);
});

// A brute force search for problems with incremental rendering...
QUnit.test("incremental render - lots of random documents", function (assert) {
	var style = fetchTextSync('external/apa.csl');
	var citeprocReact = newCiteprocReact();

	// lots of random clusterLists: compare
	var randomSeed = 0;

	var NUM_DOCUMENTS = 20, // increase significantly for better coverage
		MAX_CLUSTERS_PER_DOCUMENT = 100,
		MAX_ITEMS_PER_CLUSTER = 6;
	
	var randomClusterList = function () {
		return _.map(_.range(psuedoRandomInteger(MAX_CLUSTERS_PER_DOCUMENT)), function (cluster) {
			var citationItems = _.map(_.range(1 + psuedoRandomInteger(MAX_ITEMS_PER_CLUSTER - 1)), function (item) {
				return testCitationItem(psuedoRandomInteger(100));
			});

			return {
				citationItems: citationItems,
				properties: {},
				citationID: "this is ignored"
			};
		});
	};

	var bunchOfRandomClusterLists = _.map(_.range(NUM_DOCUMENTS), function () {
		return randomClusterList();
	});

	var originalResults = [];

	// Run in reset mode - full re-render each time
	console.time('random documents - reset mode');
	_.each(bunchOfRandomClusterLists, function (clusters, index) {
		var result = citeprocReact.render({
			style: style,
			citationClusters: clusters,
			reset: true
		});
	
		originalResults.push(_.omit(result, 'info'));
	});
	console.timeEnd('random documents - reset mode');

	// Re-run in incremental mode and compare to original
	console.time('random documents - inc mode');
	_.each(bunchOfRandomClusterLists, function (clusters, index) {
		var result = citeprocReact.render({
			style: style,
			citationClusters: clusters
		});
	
		assert.deepEqual(_.omit(result, 'info'), originalResults[index]);
	});
	console.timeEnd('random documents - inc mode');
});

var benchmark = function (styleName, assert, alwaysReset, time) {
	var style = fetchTextSync('external/' + styleName + '.csl');
	var citeprocReact = newCiteprocReact();

	var testClusters = JSON.parse(fetchTextSync('testCitations.json'));
	var testItems = JSON.parse(fetchTextSync('testItems.json'));

	// set item keys correctly for bibliography test
	_.each(testItems, function (item, key) {
		item.id = key;
	});

	var result1 = citeprocReact.render({
		style: style,
		citationClusters: testClusters,
		items: testItems,
		reset: alwaysReset
	});

	// Alter by appending and prepending a citation cluster
	testClusters.splice(0, 0, {citationItems: [testCitationItem(1), testCitationItem(2)]});

	var renderType = alwaysReset ? "full" : "incremental";

	time("prepend - " + styleName + " - " + renderType, function () {
		var result3 = citeprocReact.render({
			style: style,
			citationClusters: testClusters,
			items: testItems,
			reset: alwaysReset
		});
	});

	testClusters.push({citationItems: [testCitationItem(3), testCitationItem(4)]});

	var result4;
	time("append - " + styleName + " - " + renderType, function () {
		result4 = citeprocReact.render({
			style: style,
			citationClusters: testClusters,
			items: testItems,
			reset: alwaysReset
		});
	});

	assert.equal(result4.formattedCitations.length, result1.formattedCitations.length + 2);
	assert.equal(result4.formattedBibliography.length, result1.formattedBibliography.length + 4);
};

QUnit.test("Benchmarking test", function (assert) {
	var results = {};

	var time = function (description, callback) {
		var startTime = new Date().getTime();
		callback();
		var totalTime = new Date().getTime() - startTime;

		results[description] = results[description] || [];
		results[description].push(totalTime);

		console.log(description + ": " + totalTime + "ms");
	};

	benchmark('apa', assert, true, time);
	benchmark('apa', assert, false, time);
	benchmark('apa', assert, true, time);
	benchmark('apa', assert, false, time);
	benchmark('apa', assert, true, time);
	benchmark('apa', assert, false, time);

	benchmark('ieee', assert, true, time);
	benchmark('ieee', assert, false, time);
	benchmark('ieee', assert, true, time);
	benchmark('ieee', assert, false, time);
	benchmark('ieee', assert, true, time);
	benchmark('ieee', assert, false, time);

	console.log('results: ', results);

	var averagedResults = _.map(results, function (times, description) {
		var average = 0;
		_.each(times, function (value) {
			average += value;
		});
		average /= times.length;
		return {
			description: description,
			time: average
		};
	});

	console.log('averaged results: ', averagedResults);
});
*/
